// Youssef El merabet

//JSEG unsupervised segmentation

/*
[Deng 2001] Y. Deng et B.S Manjunath. Unsupervised Segmentation of Color-Texture
Regions in Images and video. IEEE Trans. Pattern Anal. Mach. Intell, vol. 23,
pages 800–810, 2001

*/

#include <cstdio>
#include <cstdlib>

#include "Jseg/segment.h"
#include "Jseg/imgutil.h"
#include "Jseg/mathutil.h"
#include "Jseg/quan.h"
#include "Jseg/memutil.h"


#define R    0
#define G    1
#define B    2

#ifdef MAIN
#undef MAIN
#include "segmentation/plabeling.cpp"
#define MAIN
#else

#include "segmentation/plabeling.cpp"

#endif

#include <pandore.h>

using namespace pandore;


template<typename T>
void convertPandoreToRGBTab(unsigned char *RGBTab, const Imx3d <T> &ims) {
    int numBands = ims.Bands();
    Ulong nb;

    for (int b = 0; b < numBands; ++b) {
        T *ps = ims.Vector(b);
        nb = 0;
        const T *pf = ims.Vector(b) + ims.VectorSize();   //// ims.VectorSize()= Number of pixels per band
        for (; ps < pf; ps++) {
            RGBTab[nb * numBands + b] = (Uchar) * ps;
            nb++;
        }
    }
}

void convertLabelToPandore(unsigned char *RGBTab, Reg2d &imd) {
    int nb, numBands = imd.Bands();  //numBands=1;
    Reg2d::ValueType *prgd = imd.Vector();

    nb = 0;
    for (int y = 0; y < imd.Height(); y++) {
        for (int x = 0; x < imd.Width(); x++) {
            *prgd++ = RGBTab[nb * numBands];
            nb++;
        }
    }
}


template<typename T>
void convertLabelToPandore(unsigned char *RGBTab, Imx3d <T> &ims) {
    int numBands = ims.Bands();
    Ulong nb;

    T *pdr = ims.Vector(R);
    T *pdg = ims.Vector(G);
    T *pdb = ims.Vector(B);

    nb = 0;
    for (int y = 0; y < ims.Height(); y++) {
        for (int x = 0; x < ims.Width(); x++) {
            *pdr++ = RGBTab[nb * numBands + R];
            *pdg++ = RGBTab[nb * numBands + G];
            *pdb++ = RGBTab[nb * numBands + B];
            nb++;
        }
    }
}


template<typename T>
Errc PJseg(const Imx3d <T> &ims, Reg2d &rgd, Imx3d <T> &imd, float threshcolor, float TQUAN) {

    float *LUV, **cb;
    int numBands, N, TR, imgsize, mapsize, l, NY, NX, NSCALE;
    unsigned char *RGBTab, *cmap, *rmap;
    unsigned char *result;

    NSCALE = -1; // number of scales
    numBands = ims.Bands();
    NY = ims.Height();
    NX = ims.Width();
    mapsize = NY * NX;
    imgsize = mapsize * numBands;

    RGBTab = (unsigned char *) malloc(imgsize * sizeof(unsigned char));
    cmap = (unsigned char *) calloc(mapsize, sizeof(unsigned char));
    rmap = (unsigned char *) calloc(mapsize, sizeof(unsigned char));
    result = (unsigned char *) malloc(imgsize * sizeof(unsigned char));
    cb = (float **) fmatrix(256, numBands);
    LUV = (float *) malloc(imgsize * sizeof(float));

    convertPandoreToRGBTab(RGBTab, ims);

    if (numBands == 3) {
        rgb2luv(RGBTab, LUV, imgsize);
    } else if (numBands == 1) {
        for (l = 0; l < imgsize; l++) {
            LUV[l] = RGBTab[l];
        }
    }
    N = quantize(LUV, cb, 1, NY, NX, numBands, TQUAN);

    if (numBands == 3) {
        rgb2luv(RGBTab, LUV, imgsize);
    } else if (numBands == 1) {
        for (l = 0; l < imgsize; l++) {
            LUV[l] = RGBTab[l];
        }
    }

    getcmap(LUV, cmap, cb, mapsize, numBands, N);
    TR = segment(rmap, cmap, N, 1, NY, NX, NSCALE);
    TR = merge1(rmap, cmap, N, 1, NY, NX, TR, threshcolor);

    //to produce an image with boundary
    result = outputEdge(RGBTab, rmap, NY, NX, numBands);
    convertLabelToPandore(result, imd);

    Reg2d imi(ims.Props());
    convertLabelToPandore(rmap, imi);
    Errc res = PLabeling(imi, rgd, 8);

    return res;
}

#ifdef MAIN
#define	USAGE	"usage: %s threshold q [-m mask] [im_in|-] [im_out|-] [rg_out|-] "
#define	PARC	2
#define	FINC	1
#define	FOUTC	2
#define	MASK	3

int main( int argc, char *argv[] ) {
   Errc result;              // The result code of the execution.
   Pobject* mask;            // The region map.
   Pobject* objin[FINC+1];   // The input objects.
   Pobject* objs[FINC+1];    // The source objects masked.
   Pobject* objout[FOUTC+1]; // The output object.
   Pobject* objd[FOUTC+1];   // The result object of the execution.
   char* parv[PARC+1];       // The input parameters.
   
   ReadArgs(argc, argv, PARC, FINC, FOUTC, &mask, objin, objs, objout, objd, parv, USAGE, MASK);

   if (objs[0]->Type() == Po_Imc2duc) {
      Imc2duc* const ims = (Imc2duc*)objs[0];
      objd[0]=new Reg2d(ims->Props());
      Reg2d* const imd=(Reg2d*)objd[0];
      
      objd[1] = new Imc2duc(ims->Size());
      Imc2duc* const imf = (Imc2duc*)objd[1];


      result = PJseg(*ims, *imd, *imf, (float)atof(parv[0]), (float)atof(parv[1]));
      goto end;
   } 
   if (objs[0]->Type() == Po_Img2duc) {
      Img2duc* const ims = (Img2duc*)objs[0];
      objd[0]=new Reg2d(ims->Props());
      Reg2d* const imd=(Reg2d*)objd[0];

      objd[1] = new Img2duc(ims->Size());
      Img2duc* const imf = (Img2duc*)objd[1];

      result = PJseg(*ims, *imd, *imf, (float)atof(parv[0]), (float)atof(parv[1]));
      goto end;
   }   
   {
      PrintErrorFormat(objin, FINC);
      result = FAILURE;
   }	

 end:
   if (result) {
      WriteArgs(argc, argv, PARC, FINC, FOUTC, &mask, objin, objs, objout, objd, MASK);
   }
   Exit(result);
   return 0;
}
#endif
